#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <curses.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <modules.h>
#include <util.h>
#include <dstring.h>
#include <interp.h>
#include <limits.h>
#include "version.h"
/// \section[NAME]
// \em[ncurses] -- a module for NSL (see \ref[sl(1)])

/// \section[OVERVIEW]

static module *this_mod;
extern procstack *get_pstack(void);

void ncurses_LTX_init(module *this)
{
	this_mod = this;
}

/*
static master_instruction_src ncurses_src[] = {
										 {NULL}
										};
*/										
static master_instruction_src ncurses_src_add[] = {
										{NULL}
										};

static int ncurses_inited;
static void cleanup_ncurses(void)
{
	endwin();
}

static stream *win_open(interpreter *i, char *name, char *arg, int writable);

static void win_close(interpreter *i, stream *s)
{
	delwin(s->data);
}

static html_chunk *win_read(interpreter *i, stream *s)
{
	static html_chunk res;
	int ch = wgetch(s->data);
	if(ch == ERR) return NULL;
	release_html_chunk(&res);
	if(ch <= UCHAR_MAX)
		set_html_char(&res, ch);
	else
	{
		const char *name = keyname(ch);
		set_tag_name(&res, ds_fix(ds_create("KEY")));
		if(name && *name)
			set_html_param(&res, ds_create("NAME"), ds_create(name));
		set_html_param(&res, ds_create("CODE"), ds_fromint(ch, 10, 0));
	}
	return &res;
}
	
static void win_writestr(interpreter *i, stream *s, char *str)
{
	waddstr(s->data, str);
}

static int map_acs_symbol(char *name)
{
	if(!name) return 0;
	if(!strcasecmp(name, "ULCORNER")) return ACS_ULCORNER;
	else if(!strcasecmp(name, "LLCORNER")) return ACS_LLCORNER;
	else if(!strcasecmp(name, "URCORNER")) return ACS_URCORNER;
	else if(!strcasecmp(name, "LRCORNER")) return ACS_LRCORNER;
	else if(!strcasecmp(name, "LTEE")) return ACS_LTEE;
	else if(!strcasecmp(name, "RTEE")) return ACS_RTEE;
	else if(!strcasecmp(name, "BTEE")) return ACS_BTEE;
	else if(!strcasecmp(name, "TTEE")) return ACS_TTEE;
	else if(!strcasecmp(name, "HLINE")) return ACS_HLINE;
	else if(!strcasecmp(name, "VLINE")) return ACS_VLINE;
	else if(!strcasecmp(name, "PLUS")) return ACS_PLUS;
	else if(!strcasecmp(name, "S1")) return ACS_S1;
	else if(!strcasecmp(name, "S9")) return ACS_S9;
	else if(!strcasecmp(name, "DIAMOND")) return ACS_DIAMOND;
	else if(!strcasecmp(name, "CKBOARD")) return ACS_CKBOARD;
	else if(!strcasecmp(name, "DEGREE")) return ACS_DEGREE;
	else if(!strcasecmp(name, "PLMINUS")) return ACS_PLMINUS;
	else if(!strcasecmp(name, "BULLET")) return ACS_BULLET;
	else if(!strcasecmp(name, "LARROW")) return ACS_LARROW;
	else if(!strcasecmp(name, "RARROW")) return ACS_RARROW;
	else if(!strcasecmp(name, "DARROW")) return ACS_DARROW;
	else if(!strcasecmp(name, "UARROW")) return ACS_UARROW;
	else if(!strcasecmp(name, "BOARD")) return ACS_BOARD;
	else if(!strcasecmp(name, "LANTERN")) return ACS_LANTERN;
	else if(!strcasecmp(name, "BLOCK")) return ACS_BLOCK;
	else if(!strcasecmp(name, "S3")) return ACS_S3;
	else if(!strcasecmp(name, "S7")) return ACS_S7;
	else if(!strcasecmp(name, "LEQUAL")) return ACS_LEQUAL;
	else if(!strcasecmp(name, "GEQUAL")) return ACS_GEQUAL;
	else if(!strcasecmp(name, "PI")) return ACS_PI;
	else if(!strcasecmp(name, "NEQUAL")) return ACS_NEQUAL;
	else if(!strcasecmp(name, "STERLING")) return ACS_STERLING;
	else return 0;
}

typedef struct key_map
{
	int key;
	char *name;
} key_map;

static key_map keys[] = {};

static const char *key_to_name(int key)
{
	static char buf[8];
	key_map *iter = keys;
	if(key >= KEY_F(0) && key <= KEY_F(63))
	{
		sprintf(buf, "F%d", key - KEY_F0);
		return buf;
	}
	for(; iter->name; iter++)
	{
		if(iter->key == key)
			return iter->name;
	}
	return "";
}

static int name_to_key(const char *name)
{
	key_map *iter = keys;
	if((*name == 'F' || *name == 'f') && isdigit(name[1]))
		return KEY_F0 + strtoul(name + 1, NULL, 10);
	for(; iter->name; iter++)
	{
		if(!strcasecmp(iter->name, name))
			return iter->key;
	}
	return 0;
}

static void win_writetag(interpreter *i, stream *s, html_chunk *ch)
{
	if(ch->type == hct_char)
		waddch(s->data, get_html_char(ch));
	else if(ch->type == hct_tag)
	{
		char *name = get_tag_name(ch);
		if(!strcasecmp(name, "BORDER"))
			box((WINDOW *)s->data, 0, 0);
		else if(!strcasecmp(name, "MOVE"))
		{
			int x, y;
			char *tmp;
			getyx((WINDOW *)s->data, y, x);
			tmp = get_html_param(ch, "X");
			if(tmp)
				x = strtoul(tmp, NULL, 10);
			else if((tmp = get_html_param(ch, "DX")))
				x += strtol(tmp, NULL, 10);
			tmp = get_html_param(ch, "Y");
			if(tmp)
				y = strtoul(tmp, NULL, 10);
			else if((tmp = get_html_param(ch, "DY")))
				y += strtol(tmp, NULL, 10);	
			wmove((WINDOW *)s->data, y, x);
		}
		else if(!strcasecmp(name, "HLINE"))
		{
			char *tmp = get_html_param(ch, "LEN");
			if(!tmp)
				recover_error("ncurses: HLINE must have LEN");
			whline((WINDOW *)s->data, ACS_HLINE, strtoul(tmp, NULL, 10));
		}
		else if(!strcasecmp(name, "VLINE"))
		{
			char *tmp = get_html_param(ch, "LEN");
			if(!tmp)
				recover_error("ncurses: VLINE must have LEN");
			wvline((WINDOW *)s->data, ACS_VLINE, strtoul(tmp, NULL, 10));
		}
		else if(!strcasecmp(name, "STANDOUT"))
			wattron((WINDOW *)s->data, A_STANDOUT);
		else if(!strcasecmp(name, "/STANDOUT"))
			wattroff((WINDOW *)s->data, A_STANDOUT);
		else if(!strcasecmp(name, "BOLD"))
			wattron((WINDOW *)s->data, A_BOLD);
		else if(!strcasecmp(name, "/BOLD"))
			wattroff((WINDOW *)s->data, A_BOLD);
		else if(!strcmp(name, "&"))
		{
			int acs = map_acs_symbol(get_html_param(ch, "VALUE"));
			if(acs)
				waddch((WINDOW *)s->data, acs);
		}
	}	
}

static char *win_ioctl(interpreter *i, stream *s, char *ctl)
{
	if(!strcasecmp(ctl, "flush"))
	{
		wrefresh(s->data);
		return "";
	}
	else if(!strcasecmp(ctl, "scrollok"))
	{
		scrollok((WINDOW *)s->data, TRUE);
		return "";
	}
	else if(!strcasecmp(ctl, "noscrollok"))
	{
		scrollok((WINDOW *)s->data, FALSE);
		return "";
	}
	return NULL;
}

static int win_getpos(interpreter *i, stream *s)
{
	int wdt, hgt, y, x;
	getmaxyx((WINDOW *)s->data, wdt, hgt);
	wdt++;
	getyx((WINDOW *)s->data, y, x);
	return y * wdt + x;
}

static void win_setpos(interpreter *i, stream *s, int pos)
{
	int wdt, hgt;
	getmaxyx((WINDOW *)s->data, wdt, hgt);
	wdt++;
	wmove((WINDOW *)s->data, pos / wdt, pos % wdt);
}

static void win_rewind(interpreter *i, stream *s)
{
	wmove(s->data, 0, 0);
}

static stream_class ncurses_stream_class = {"window", win_open, win_close,
											win_read, win_writestr, win_writetag,
											win_ioctl, win_getpos, win_setpos, win_rewind
										};

static stream *win_open(interpreter *i, char *name, char *arg, int writable)
{
	stream *ns = ALLOC(stream);
	WINDOW *nw;
	int wdt, hgt, x, y;
	ns->class = &ncurses_stream_class;
	ns->direction = 1;
	wdt = strtol(arg, &arg, 10);
	if(*arg == 'x')
		arg++;
	hgt = strtol(arg, &arg, 10);
	x = strtol(arg, &arg, 10);
	y = strtol(arg, &arg, 10);
	if(x < 0)
		x = COLS  - wdt + x + 1;
	if(y < 0)
		y = LINES - hgt + y + 1; 
	nw = newwin(hgt, wdt, y, x);
	if(!nw)
		recover_error("ncurses error: can't create a window");
	keypad(nw, TRUE);
	idlok(nw, TRUE);
	ns->data = nw;
	return ns;
}

void ncurses_LTX_attach(mod_attach *ma)
{
	if(!ncurses_inited)
	{
		initscr();
		atexit(cleanup_ncurses);
		if(has_colors())
			start_color();
		cbreak();
		noecho();
		nonl();
		intrflush(stdscr, FALSE);
		ncurses_inited = 1;
	}	
	register_mod_stream(ma->i, &ncurses_stream_class);
//	ma->data = xmalloc(sizeof(ncurses_src) * sizeof(instruction_t) / sizeof(*ncurses_src));
//	replace_builtins(ma->i, ncurses_src, ma->data);
	append_builtins(ma->i, ncurses_src_add);
}

const char ncurses_LTX_version[] = VERSION;
const char ncurses_LTX_date[] = __DATE__;

/// \section[AUTHOR]

/// \section[COPYRIGHT]
// see \ref[sl(1)]

/// \seealso
